-- YGREC8/Core.vhdl
-- created dim. nov. 12 22:26:45 CET 2017 by Yann Guidon (whygee@f-cpu.org)
-- version mar. nov. 14 03:10:08 CET 2017
-- version mar. nov. 14 04:34:04 CET 2017 : yeah ! datapath written !
-- Released under the GNU AGPLv3 license
--
-- This is the main file of YGREC8,
-- it assembles the sub-units to make the core's datapath.
-- Many signals and registers are set as inout mode to ease coding.

Library ieee;
    use ieee.std_logic_1164.all;
Library work;
    use work.all;
    use work.ygrec8_def.all;
    use work.ygrec8_utils.all;

entity YGREC8_Core is
  port(
    clk, reset, enable_cycle,
    CB0, CB1, CB2, CB3 : in std_logic;
    InstructionWord : in std_logic_vector(15 downto 0);
    IN_PORT : std_logic_vector(7 downto 0);
    ADR_PORT : out std_logic_vector(8 downto 0);
    PC, NextPC, WriteBack, DST, SRC : inout std_logic_vector(7 downto 0);
    nC, C, nS, S, nZ, Z, write_enable : inout std_logic);
 -- OUT_PORT : use DST instead
end YGREC8_Core;

architecture A3P of YGREC8_Core is
  signal AdrSRC, AdrDST, Condition :  std_logic_vector(2 downto 0);
  signal Opcode : std_logic_vector(3 downto 0);
  signal Imm8, Imm3ext, PC1, SRC_reg, SHL, SUM, ROP2, Result,
      D1, D2, A1, A2 : std_logic_vector(7 downto 0);
  signal RI3, RI8, Cin, Cout, disable_write,
    a_wr_en, b_wr_en, CondNeg, CondSel : std_logic;
  signal a_addr, b_addr : std_logic_vector(8 downto 0);
begin
  -- alias the instruction fields to the instruction word :
  Imm8      <= InstructionWord( 7 downto  0);
  AdrSRC    <= InstructionWord( 2 downto  0);
  Imm3ext   <= (    0=>InstructionWord(0),
                    1=>InstructionWord(1),
               others=>InstructionWord(2));
  RI3       <= InstructionWord( 3);
  Condition <= InstructionWord( 6 downto  4);
  CondNeg   <= InstructionWord( 7);
  RI8       <= InstructionWord( 8);
  AdrDST    <= InstructionWord(11 downto  9);
  Opcode    <= InstructionWord(15 downto 12);
  ADR_PORT  <= InstructionWord( 8 downto  0);

  disable_write <= '1' when
             Opcode=Op_OUT
          OR Opcode=Op_CMPU
          OR Opcode=Op_CMPS
          OR enable_cycle='0'
    else '0';

  entity_INC_PC: entity INC8 port map (
     A => PC,
     Y => PC1,
     V => open);  -- => used to trigger a trap later

  WriteBack <= PC when Opcode=Op_CALL
          else Result;
  NextPC <= Result when (Opcode=Op_CALL)
                     or (AdrDST=Reg_PC and disable_write='0')
    else PC1;

  entity_REG8: entity REG8 port map (
    clk => clk,
    reset => reset,
    write_enable => write_enable,
    AdrSRC => AdrSRC,
    AdrDST => AdrDST,
    PC => PC,
    D1 => D1,
    D2 => D2,
    RESULT => WriteBack,
    SRC => SRC_reg,
    DST => DST,
    A1 => A1,
    A2 => A2
  );

--  OUT_PORT <= DST;
  SRC <= SRC_reg when (RI8='0' and RI3='1')  -- slow path first
    else Imm8  when RI8='1'                  -- fast lane next
    else Imm3ext; -- (RI8='0' and RI3='0')

  Cin <= '0' when Opcode(1 downto 0)=Op_ADD(1 downto 0)
    else '1';  -- no ADDC or SUBB support. yet ?

  entity_ALU8: entity ALU8 port map (
     F0 => Opcode(0),
     F1 => Opcode(1),
     F2 => Opcode(2),
     F3 => Opcode(3),
     Cin => Cin,
     SRC => SRC,
     DST => DST,
     SUM => SUM,
     ROP2 => ROP2,
     Cout => Cout);

  entity_SHL8: entity SHL8 port map (
     F0 => Opcode(0),
     F1 => Opcode(1),
     F2 => Opcode(2),
     F3 => Opcode(3),
     SRC => SRC,
     DST => DST,
     SHL => SHL);

  Result <= SUM  when Opcode(3 downto 2)=Op_ADD(3 downto 2) -- ADD, SUB, CMPU, CMPS
       else SHL  when Opcode(3 downto 2)=Op_SHL(3 downto 2) -- SHR, SHL, SAR,  ROL
       else IN_PORT when Opcode=Op_IN                       -- IN (OUT does not write)
       else ROP2 when Opcode(3 downto 2)=Op_OR (3 downto 2) -- OR, XOR, AND, ANDN
       else SRC;                                            -- MOV, CALL

  nC <= Cout;
  nS <= Result(7);
  nZ <= '0' when Result="00000000"
   else '1';

  CondSel <= '1' when Condition=Cond_A
        else  C  when Condition=Cond_C
        else  S  when Condition=Cond_S
        else  Z  when Condition=Cond_Z
        else CB0 when Condition=Cond_0
        else CB1 when Condition=Cond_1
        else CB2 when Condition=Cond_2
        else CB3 ; -- Condition=Cond_3
  write_enable <= CondSel xor CondNeg when disable_write='0' 
            else '0';

  -- The register set has its own registers,
  -- this process only has to deal with the flags and PC.
  process (clk, reset)
  begin
    if reset='0' then
       C <= '0';
       S <= '0';
       Z <= '0';
      PC <= "00000000";
    else
      if rising_edge(clk) then
        if enable_cycle='1' then
          -- advance the program:
          PC <= NextPC;

          if write_enable='1' then
            -- update only if DST register is written
            S <= nS;
            Z <= nZ;
            if Opcode(3 downto 2)=Op_ADD(3 downto 2) then
              -- Update Carry only if ADD, SUB, CMPU, CMPS
              C <= nC;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  -- Memory system:

  -- The MSB should be configurable somehow under program control...
  a_addr(8) <= '0';
  b_addr(8) <= '1';

  a_addr(7 downto 0) <= WriteBack when AdrDST=Reg_A1 and write_enable='1'
       else A1;
  b_addr(7 downto 0) <= WriteBack when AdrDST=Reg_A2 and write_enable='1'
       else A2;

  a_wr_en <= '0' when AdrDST=Reg_D1 and write_enable='1'
        else '1';
  b_wr_en <= '0' when AdrDST=Reg_D2 and write_enable='1'
        else '1';

  entity_bram: entity bram 
  generic map(
    SLICES      => 1,
    SLICE_WIDTH => 8,
    WIDTH       => 8, -- please preserve SLICES*SLICE_WIDTH;
    ADRBITS     => 9
    -- For Microsemi :
    -- ADRBITS       9  10  11  12
    -- SLICE_WIDTH  8/9  4   2   1
  )
  port map (
    reset   => reset,
    -- port A :
    a_clk   => clk,
    a_wr_en => a_wr_en,
    a_addr  => a_addr,
    a_din   => WriteBack,
    a_dout  => D1,
    -- port B :
    b_clk   => clk,
    b_wr_en => b_wr_en,
    b_addr  => b_addr,
    b_din   => WriteBack,
    b_dout  => D2 );

end A3P;
